/*
 * "$Id: form-ps.c,v 1.1.1.9 2004/06/05 02:42:32 jlovell Exp $"
 *
 *   CUPS form PostScript routines for the Common UNIX Printing System (CUPS).
 *
 *   Copyright 1997-2004 by Easy Software Products.
 *
 *   These coded instructions, statements, and computer programs are the
 *   property of Easy Software Products and are protected by Federal
 *   copyright law.  Distribution and use rights are outlined in the file
 *   "LICENSE.txt" which should have been included with this file.  If this
 *   file is missing or damaged please contact Easy Software Products
 *   at:
 *
 *       Attn: CUPS Licensing Information
 *       Easy Software Products
 *       44141 Airport View Drive, Suite 204
 *       Hollywood, Maryland 20636-3111 USA
 *
 *       Voice: (301) 373-9603
 *       EMail: cups-info@cups.org
 *         WWW: http://www.cups.org
 *
 *   This file is subject to the Apple OS-Developed Software exception.
 *
 * Contents:
 *
 */

/*
 * Include necessary headers...
 */

#include "form.h"


/*
 * 'formWrite()' - Write PostScript output for the given form document.
 */

void
formWrite(tree_t *t)			/* I - Document tree to write */
{
}


/*
 * End of "$Id: form-ps.c,v 1.1.1.9 2004/06/05 02:42:32 jlovell Exp $".
 */
